var GomokuDirectionType;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(GomokuDirectionType || (GomokuDirectionType = {}));
const gomokuDirections = Array.from(Object.entries({
    [GomokuDirectionType.TOP]: [-1, 0],
    [GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [GomokuDirectionType.RIGHT]: [0, 1],
    [GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [GomokuDirectionType.BOTTOM]: [1, 0],
    [GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [GomokuDirectionType.LEFT]: [0, -1],
    [GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const leftHalfGomokuDirectionTypes = [
    GomokuDirectionType.LEFT,
    GomokuDirectionType.TOP_LEFT,
    GomokuDirectionType.TOP,
    GomokuDirectionType.TOP_RIGHT,
];
const rightHalfGomokuDirectionTypes = [
    GomokuDirectionType.RIGHT,
    GomokuDirectionType.BOTTOM_RIGHT,
    GomokuDirectionType.BOTTOM,
    GomokuDirectionType.BOTTOM_LEFT,
];
const gomokuDirectionTypes = leftHalfGomokuDirectionTypes.concat(rightHalfGomokuDirectionTypes);

class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_INLINE;
    TOTAL_POS;
    TOTAL_PLAYERS;
    board;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE) {
        this.MAX_ROW = MAX_ROW;
        this.MAX_COL = MAX_COL;
        this.MAX_INLINE = MAX_INLINE;
        this.TOTAL_POS = MAX_ROW * MAX_COL;
        this.TOTAL_PLAYERS = 2;
        this.board = new Int32Array(this.TOTAL_POS);
    }
    idx(r, c) {
        return r * this.MAX_ROW + c;
    }
    idxIfValid(r, c) {
        return this.isValidPos(r, c) ? r * this.MAX_ROW + c : -1;
    }
    reIdx(id) {
        const c = id % this.MAX_ROW;
        const r = Math.round((id - c) / this.MAX_ROW);
        return [r, c];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    isInvalidPos(r, c) {
        return r < 0 || r >= this.MAX_ROW || c < 0 || c >= this.MAX_COL;
    }
    move(r, c, dirType, step = 1) {
        const [dr, dc] = gomokuDirections[dirType];
        const r2 = r + dr * step;
        const c2 = c + dc * step;
        return [r2, c2];
    }
    visitValidNeighbors(r, c, touch) {
        for (const dirType of gomokuDirectionTypes) {
            const [r2, c2] = this.move(r, c, dirType);
            if (this.isValidPos(r2, c2))
                touch(r2, c2, dirType);
        }
    }
    hasPlacedNeighbors(r, c) {
        const { board } = this;
        for (const dirType of gomokuDirectionTypes) {
            const [r2, c2] = this.move(r, c, dirType);
            if (this.isValidPos(r2, c2)) {
                const id2 = this.idx(r2, c2);
                if (board[id2] >= 0)
                    return true;
            }
        }
        return false;
    }
    traverseAllDirections(handle) {
        this.traverseLeftDirections(handle);
        this.traverseRightDirections(handle);
    }
    traverseLeftDirections(handle) {
        const { MAX_ROW, MAX_COL } = this;
        for (const dirType of leftHalfGomokuDirectionTypes) {
            for (let r = 0; r < MAX_ROW; ++r) {
                for (let c = 0; c < MAX_COL; ++c)
                    handle(r, c, dirType);
            }
        }
    }
    traverseRightDirections(handle) {
        const { MAX_ROW, MAX_COL } = this;
        const R = MAX_ROW - 1;
        const C = MAX_COL - 1;
        for (const dirType of rightHalfGomokuDirectionTypes) {
            for (let r = R; r >= 0; --r) {
                for (let c = C; c >= 0; --c)
                    handle(r, c, dirType);
            }
        }
    }
}

class GomokuCountMap {
    context;
    dirCountMap;
    conShapeCountMap;
    gapShapeCountMap;
    scoreMap;
    NEXT_MOVER_FAC;
    constructor(context, scoreMap, NEXT_MOVER_FAC = 1.4) {
        this.context = context;
        this.dirCountMap = new Array(gomokuDirections.length)
            .fill([])
            .map(() => new Uint32Array(context.TOTAL_POS));
        this.conShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.gapShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.scoreMap = scoreMap;
        this.NEXT_MOVER_FAC = NEXT_MOVER_FAC;
    }
    init() {
        this.dirCountMap.forEach(idMap => idMap.fill(0));
        this.conShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        this.gapShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        const { context, dirCountMap } = this;
        const { board } = context;
        context.traverseAllDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const player = board[id];
            if (player >= 0) {
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
        });
        context.traverseLeftDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const player = board[id];
            if (player >= 0 && dirCountMap[dirType][id] === 1) {
                const revDirType = dirType ^ 1;
                const cnt = dirCountMap[revDirType][id];
                this.updateConShapeCountMap(player, r, c, cnt, revDirType, 1);
            }
        });
        context.traverseLeftDirections((r, c, dirType) => {
            const id = context.idx(r, c);
            const revDirType = dirType ^ 1;
            if (board[id] < 0) {
                const [r0, c0] = context.move(r, c, revDirType, 1);
                const id0 = context.idxIfValid(r0, c0);
                if (id0 < 0 || board[id0] < 0)
                    return;
                const player = board[id0];
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                if (id2 < 0 || board[id2] !== player)
                    return;
                const [cnt0, countOfFreeSide0] = this.detectFreeSide(r0, c0, revDirType);
                const [cnt2, countOfFreeSide2] = this.detectFreeSide(r2, c2, dirType);
                const countOfFreeSide = countOfFreeSide0 + countOfFreeSide2;
                this.updateGapShapeCountMap(player, cnt0, cnt2, countOfFreeSide, 1);
            }
        });
    }
    beforeForward(r, c) {
        const { context } = this;
        const { board } = context;
        for (const dirType of gomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, 1, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const [r0, c0] = context.move(r, c, revDirType, 1);
            const id0 = context.idxIfValid(r0, c0);
            const [r2, c2] = context.move(r, c, dirType, 1);
            const id2 = context.idxIfValid(r2, c2);
            if (id0 >= 0 && id2 >= 0 && board[id0] === board[id2] && board[id0] >= 0) {
                const player0 = board[id0];
                const [cnt0, countOfFreeSide0] = this.detectFreeSide(r0, c0, revDirType);
                const [cnt2, countOfFreeSide2] = this.detectFreeSide(r2, c2, dirType);
                const countOfFreeSide = countOfFreeSide0 + countOfFreeSide2;
                this.updateGapShapeCountMap(player0, cnt0, cnt2, countOfFreeSide, -1);
            }
            if (id0 >= 0 && board[id0] >= 0) {
                const player0 = board[id0];
                const gapShape = this.detectGapContinuousLine(r0, c0, player0, revDirType);
                if (gapShape) {
                    const [lftCnt, rhtCnt, countOfFreeSide] = gapShape;
                    this.updateGapShapeCountMap(player0, lftCnt, rhtCnt, countOfFreeSide, -1);
                }
            }
            if (id2 >= 0 && board[id2] >= 0) {
                const player2 = board[id2];
                const gapShape = this.detectGapContinuousLine(r2, c2, player2, dirType);
                if (gapShape) {
                    const [lftCnt, rhtCnt, countOfFreeSide] = gapShape;
                    this.updateGapShapeCountMap(player2, lftCnt, rhtCnt, countOfFreeSide, -1);
                }
            }
        }
    }
    afterForward(r, c) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const id = context.idx(r, c);
        const player = board[id];
        {
            for (const dirType of gomokuDirectionTypes) {
                const [r2, c2] = context.move(r, c, dirType, 1);
                const id2 = context.idxIfValid(r2, c2);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
            for (const dirType of gomokuDirectionTypes) {
                const [dr, dc] = gomokuDirections[dirType];
                const countMap = dirCountMap[dirType];
                const cnt = countMap[id];
                for (let r2 = r, c2 = c;;) {
                    r2 = r2 - dr;
                    c2 = c2 - dc;
                    if (context.isInvalidPos(r2, c2))
                        break;
                    const id2 = context.idx(r2, c2);
                    if (board[id2] !== player)
                        break;
                    countMap[id2] += cnt;
                }
            }
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const lftCnt = dirCountMap[revDirType][id];
            const rhtCnt = dirCountMap[dirType][id];
            const [startR, startC] = context.move(r, c, revDirType, lftCnt - 1);
            this.updateConShapeCountMap(player, startR, startC, lftCnt + rhtCnt - 1, dirType, 1);
            this.updateRelatedConShapeCountMap(r, c, lftCnt, revDirType, 1);
            this.updateRelatedConShapeCountMap(r, c, rhtCnt, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const gapShape0 = this.detectGapContinuousLine(r, c, player, revDirType);
            if (gapShape0) {
                const [lftCnt, rhtCnt, countOfFreeSide] = gapShape0;
                this.updateGapShapeCountMap(player, lftCnt, rhtCnt, countOfFreeSide, 1);
            }
            else {
                const [r00, c00] = context.move(r, c, revDirType, dirCountMap[revDirType][id]);
                const id00 = context.idxIfValid(r00, c00);
                if (id00 >= 0 && board[id00] >= 0) {
                    const player00 = board[id00];
                    const gapShape00 = this.detectGapContinuousLine(r00, c00, player00, revDirType);
                    if (gapShape00) {
                        const [lftCnt, rhtCnt, countOfFreeSide] = gapShape00;
                        this.updateGapShapeCountMap(player00, lftCnt, rhtCnt, countOfFreeSide, 1);
                    }
                }
            }
            const gapShape2 = this.detectGapContinuousLine(r, c, player, dirType);
            if (gapShape2) {
                const [lftCnt, rhtCnt, countOfFreeSide] = gapShape2;
                this.updateGapShapeCountMap(player, lftCnt, rhtCnt, countOfFreeSide, 1);
            }
            else {
                const [r22, c22] = context.move(r, c, dirType, dirCountMap[dirType][id]);
                const id22 = context.idxIfValid(r22, c22);
                if (id22 >= 0 && board[id22] >= 0) {
                    const player22 = board[id22];
                    const gapShape22 = this.detectGapContinuousLine(r22, c22, player22, dirType);
                    if (gapShape22) {
                        const [lftCnt, rhtCnt, countOfFreeSide] = gapShape22;
                        this.updateGapShapeCountMap(player22, lftCnt, rhtCnt, countOfFreeSide, 1);
                    }
                }
            }
        }
    }
    beforeRollback(r, c) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const id = context.idx(r, c);
        const player = board[id];
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const lftCnt = dirCountMap[revDirType][id];
            const rhtCnt = dirCountMap[dirType][id];
            const [startR, startC] = context.move(r, c, revDirType, lftCnt - 1);
            this.updateConShapeCountMap(player, startR, startC, lftCnt + rhtCnt - 1, dirType, -1);
            this.updateRelatedConShapeCountMap(r, c, lftCnt, revDirType, -1);
            this.updateRelatedConShapeCountMap(r, c, rhtCnt, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const gapShape0 = this.detectGapContinuousLine(r, c, player, revDirType);
            if (gapShape0) {
                const [lftCnt, rhtCnt, countOfFreeSide] = gapShape0;
                this.updateGapShapeCountMap(player, lftCnt, rhtCnt, countOfFreeSide, -1);
            }
            else {
                const [r00, c00] = context.move(r, c, revDirType, dirCountMap[revDirType][id]);
                const id00 = context.idxIfValid(r00, c00);
                if (id00 >= 0 && board[id00] >= 0) {
                    const player00 = board[id00];
                    const gapShape00 = this.detectGapContinuousLine(r00, c00, player00, revDirType);
                    if (gapShape00) {
                        const [lftCnt, rhtCnt, countOfFreeSide] = gapShape00;
                        this.updateGapShapeCountMap(player00, lftCnt, rhtCnt, countOfFreeSide, -1);
                    }
                }
            }
            const gapShape2 = this.detectGapContinuousLine(r, c, player, dirType);
            if (gapShape2) {
                const [lftCnt, rhtCnt, countOfFreeSide] = gapShape2;
                this.updateGapShapeCountMap(player, lftCnt, rhtCnt, countOfFreeSide, -1);
            }
            else {
                const [r22, c22] = context.move(r, c, dirType, dirCountMap[dirType][id]);
                const id22 = context.idxIfValid(r22, c22);
                if (id22 >= 0 && board[id22] >= 0) {
                    const player22 = board[id22];
                    const gapShape22 = this.detectGapContinuousLine(r22, c22, player22, dirType);
                    if (gapShape22) {
                        const [lftCnt, rhtCnt, countOfFreeSide] = gapShape22;
                        this.updateGapShapeCountMap(player22, lftCnt, rhtCnt, countOfFreeSide, -1);
                    }
                }
            }
        }
    }
    afterRollback(r, c, player) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const id = context.idx(r, c);
        {
            for (const dirType of gomokuDirectionTypes) {
                const [dr, dc] = gomokuDirections[dirType];
                const countMap = dirCountMap[dirType];
                const cnt = countMap[id];
                for (let r2 = r, c2 = c;;) {
                    r2 = r2 - dr;
                    c2 = c2 - dc;
                    const id2 = context.idxIfValid(r2, c2);
                    if (id2 < 0 || board[id2] !== player)
                        break;
                    countMap[id2] -= cnt;
                }
            }
            for (const dirType of gomokuDirectionTypes)
                dirCountMap[dirType][id] = 0;
        }
        for (const dirType of gomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(r, c, 1, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            const revDirType = dirType ^ 1;
            const [r0, c0] = context.move(r, c, revDirType, 1);
            const id0 = context.idxIfValid(r0, c0);
            const [r2, c2] = context.move(r, c, dirType, 1);
            const id2 = context.idxIfValid(r2, c2);
            if (id0 >= 0 && id2 >= 0 && board[id0] === board[id2] && board[id0] >= 0) {
                const player0 = board[id0];
                const [cnt0, countOfFreeSide0] = this.detectFreeSide(r0, c0, revDirType);
                const [cnt2, countOfFreeSide2] = this.detectFreeSide(r2, c2, dirType);
                const countOfFreeSide = countOfFreeSide0 + countOfFreeSide2;
                this.updateGapShapeCountMap(player0, cnt0, cnt2, countOfFreeSide, 1);
            }
            if (id0 >= 0 && board[id0] >= 0) {
                const player0 = board[id0];
                const gapShape = this.detectGapContinuousLine(r0, c0, player0, revDirType);
                if (gapShape) {
                    const [lftCnt, rhtCnt, countOfFreeSide] = gapShape;
                    this.updateGapShapeCountMap(player0, lftCnt, rhtCnt, countOfFreeSide, 1);
                }
            }
            if (id2 >= 0 && board[id2] >= 0) {
                const player2 = board[id2];
                const gapShape = this.detectGapContinuousLine(r2, c2, player2, dirType);
                if (gapShape) {
                    const [lftCnt, rhtCnt, countOfFreeSide] = gapShape;
                    this.updateGapShapeCountMap(player2, lftCnt, rhtCnt, countOfFreeSide, 1);
                }
            }
        }
    }
    score(currentPlayer, scoreForPlayer) {
        if (this.hasReachedTheLimit(scoreForPlayer))
            return Number.POSITIVE_INFINITY;
        if (this.hasReachedTheLimit(scoreForPlayer ^ 1))
            return Number.NEGATIVE_INFINITY;
        const { context, conShapeCountMap, gapShapeCountMap, scoreMap } = this;
        const conCountMap = conShapeCountMap[scoreForPlayer];
        const gapCountMap = gapShapeCountMap[scoreForPlayer];
        let score = 0;
        for (let cnt = 1; cnt < context.MAX_INLINE; ++cnt) {
            const [a, b, c] = conCountMap[cnt];
            const [x, y, z] = scoreMap.continuously[cnt];
            score += a * x + b * y + c * z;
        }
        for (let cnt = 1; cnt <= context.MAX_INLINE; ++cnt) {
            const [a, b, c] = gapCountMap[cnt];
            const [x, y, z] = scoreMap.gap[cnt];
            score += a * x + b * y + c * z;
        }
        return currentPlayer === scoreForPlayer ? score : score * this.NEXT_MOVER_FAC;
    }
    hasReachedTheLimit(player) {
        const { context, conShapeCountMap } = this;
        return conShapeCountMap[player][context.MAX_INLINE].some(x => x > 0);
    }
    toJSON() {
        const { dirCountMap, conShapeCountMap, gapShapeCountMap } = this;
        return { dirCountMap, conShapeCountMap, gapShapeCountMap };
    }
    updateRelatedConShapeCountMap(centerR, centerC, startCnt, dirType, v) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const [dr, dc] = gomokuDirections[dirType];
        let threshold = context.MAX_INLINE - startCnt;
        let r = centerR + dr * startCnt;
        let c = centerC + dc * startCnt;
        let player = -1;
        if (startCnt > 1) {
            const id = context.idx(centerR, centerC);
            if (id >= 0)
                player = board[id];
        }
        for (; threshold > 0;) {
            const id = context.idxIfValid(r, c);
            if (id < 0)
                break;
            if (board[id] < 0) {
                r += dr;
                c += dc;
                threshold -= 1;
                continue;
            }
            if (player >= 0 && player !== board[id])
                break;
            player = board[id];
            const cnt = dirCountMap[dirType][id];
            this.updateConShapeCountMap(player, r, c, cnt, dirType, v);
            r += dr * cnt;
            c += dc * cnt;
            threshold -= cnt;
        }
    }
    updateConShapeCountMap(player, startR, startC, cnt, dirType, v) {
        const countOfFreeSide = this.detectExpandable(player, startR, startC, cnt, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.conShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    updateGapShapeCountMap(player, lftCnt, rhtCnt, countOfFreeSide, v) {
        const threshold = this.context.MAX_INLINE - 1;
        if (lftCnt >= threshold || rhtCnt >= threshold)
            return;
        const cnt = lftCnt + rhtCnt;
        const _cnt = Math.min(cnt, this.context.MAX_INLINE);
        this.gapShapeCountMap[player][_cnt][countOfFreeSide] += v;
    }
    detectGapContinuousLine(r, c, player, dirType) {
        const revDirType = dirType ^ 1;
        const [cnt0, countOfFreeSide0] = this.detectFreeSide(r, c, revDirType);
        const [cnt2, countOfFreeSide2] = this.detectFreeSide(r, c, dirType);
        if (countOfFreeSide2) {
            const { context } = this;
            const [r22, c22] = context.move(r, c, dirType, cnt2 + 1);
            const id22 = context.idxIfValid(r22, c22);
            if (id22 >= 0 && context.board[id22] === player) {
                const [cnt22, countOfFreeSide22] = this.detectFreeSide(r22, c22, dirType);
                const countOfFreeSide = countOfFreeSide0 + countOfFreeSide22;
                return [cnt0 + cnt2 - 1, cnt22, countOfFreeSide];
            }
        }
        return null;
    }
    detectFreeSide(r, c, dirType) {
        const id = this.context.idx(r, c);
        const { context, dirCountMap } = this;
        const cnt = dirCountMap[dirType][id];
        const [r2, c2] = context.move(r, c, dirType, cnt);
        const id2 = context.idxIfValid(r2, c2);
        const result = [cnt, id2 >= 0 && context.board[id2] < 0 ? 1 : 0];
        return result;
    }
    detectExpandable(player, startR, startC, cnt, dirType) {
        const { context } = this;
        const { MAX_INLINE, board } = context;
        const revDirType = dirType ^ 1;
        const [r0, c0] = context.move(startR, startC, revDirType, 1);
        const id0 = context.idxIfValid(r0, c0);
        const isFreeSide0 = id0 >= 0 && board[id0] < 0 ? 1 : 0;
        const [r2, c2] = context.move(startR, startC, dirType, cnt);
        const id2 = context.idxIfValid(r2, c2);
        const isFreeSide2 = id2 >= 0 && board[id2] < 0 ? 1 : 0;
        const countOfFreeSide = isFreeSide0 + isFreeSide2;
        if (countOfFreeSide === 0)
            return 0;
        let maxPossibleCnt = cnt + countOfFreeSide;
        if (maxPossibleCnt < MAX_INLINE) {
            if (isFreeSide0) {
                const [dr, dc] = gomokuDirections[revDirType];
                for (let r = r0, c = c0; maxPossibleCnt < MAX_INLINE; ++maxPossibleCnt) {
                    r += dr;
                    c += dc;
                    const id = context.idxIfValid(r, c);
                    if (id < 0 || (board[id] >= 0 && board[id] !== player))
                        break;
                }
            }
            if (isFreeSide2) {
                const [dr, dc] = gomokuDirections[dirType];
                for (let r = r2, c = c2; maxPossibleCnt < MAX_INLINE; ++maxPossibleCnt) {
                    r += dr;
                    c += dc;
                    const id = context.idxIfValid(r, c);
                    if (id < 0 || (board[id] >= 0 && board[id] !== player))
                        break;
                }
            }
        }
        return maxPossibleCnt < MAX_INLINE ? 0 : countOfFreeSide;
    }
    countDangerShapes(player) {
        const { context, conShapeCountMap, gapShapeCountMap } = this;
        const shapeCountMap1 = conShapeCountMap[player][context.MAX_INLINE - 1];
        const shapeCountMap2 = gapShapeCountMap[player][context.MAX_INLINE];
        const cnt = shapeCountMap1[1] +
            shapeCountMap1[2] * 2 +
            shapeCountMap2[0] +
            shapeCountMap2[1] +
            shapeCountMap2[2];
        return cnt;
    }
}

class GomokuState {
    context;
    countMap;
    candidateSet;
    placedCount;
    constructor(context, scoreMap, NEXT_MOVER_FAC) {
        this.context = context;
        this.countMap = new GomokuCountMap(context, scoreMap, NEXT_MOVER_FAC);
        this.candidateSet = new Set();
        this.placedCount = 0;
    }
    init(pieces = []) {
        const { context, candidateSet } = this;
        const board = context.board;
        this.placedCount = pieces.length;
        board.fill(-1);
        candidateSet.clear();
        for (const { r, c, p } of pieces) {
            const id = context.idx(r, c);
            board[id] = p;
            candidateSet.delete(id);
            context.visitValidNeighbors(r, c, (r2, c2) => {
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0)
                    candidateSet.add(id2);
            });
        }
        this.countMap.init();
    }
    forward(r, c, player) {
        const { context, candidateSet } = this;
        const board = context.board;
        const id = context.idxIfValid(r, c);
        if (id < 0 || board[id] >= 0)
            return;
        this.beforeForward(r, c);
        {
            this.placedCount += 1;
            board[id] = player;
            candidateSet.delete(id);
            context.visitValidNeighbors(r, c, (r2, c2) => {
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0)
                    candidateSet.add(id2);
            });
        }
        this.afterForward(r, c);
    }
    rollback(r, c) {
        const { context, candidateSet } = this;
        const board = context.board;
        const id = context.idxIfValid(r, c);
        if (id < 0 || board[id] < 0)
            return;
        const player = board[id];
        this.beforeRollback(r, c);
        {
            this.placedCount -= 1;
            board[id] = -1;
            if (context.hasPlacedNeighbors(r, c))
                candidateSet.add(id);
            context.visitValidNeighbors(r, c, (r2, c2) => {
                const id2 = context.idx(r2, c2);
                if (board[id2] >= 0 || !context.hasPlacedNeighbors(r2, c2)) {
                    candidateSet.delete(id2);
                }
            });
        }
        this.afterRollback(r, c, player);
    }
    expand(currentPlayer, scoreForPlayer) {
        const { context, candidateSet } = this;
        const { board } = context;
        const candidates = [];
        for (const id of candidateSet) {
            if (board[id] >= 0)
                continue;
            const [r, c] = context.reIdx(id);
            candidates.push({ r, c, score: 0 });
        }
        for (const candidate of candidates) {
            const { r, c } = candidate;
            this.forward(r, c, currentPlayer);
            candidate.score = this.score(currentPlayer, scoreForPlayer);
            this.rollback(r, c);
        }
        return candidates;
    }
    score(currentPlayer, scoreForPlayer) {
        return (this.countMap.score(currentPlayer, scoreForPlayer) -
            this.countMap.score(currentPlayer, scoreForPlayer ^ 1));
    }
    isFinal() {
        const { context, countMap, placedCount } = this;
        if (placedCount === context.TOTAL_POS)
            return true;
        for (let player = 0; player < context.TOTAL_PLAYERS; ++player) {
            if (countMap.hasReachedTheLimit(player))
                return true;
        }
        return false;
    }
    randomMove() {
        const { context, candidateSet } = this;
        const { board } = context;
        let result = null;
        for (const id of candidateSet) {
            const [r, c] = context.reIdx(id);
            context.visitValidNeighbors(r, c, (r2, c2) => {
                if (result)
                    return;
                const id2 = context.idx(r2, c2);
                if (board[id2] < 0) {
                    result = { r: r2, c: c2 };
                }
            });
        }
        return result ?? { r: -1, c: -1 };
    }
    beforeForward(r, c) {
        this.countMap.beforeForward(r, c);
    }
    afterForward(r, c) {
        this.countMap.afterForward(r, c);
    }
    beforeRollback(r, c) {
        this.countMap.beforeRollback(r, c);
    }
    afterRollback(r, c, player) {
        this.countMap.afterRollback(r, c, player);
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

const createScoreMap = (MAX_INLINE) => {
    const continuously = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    const gap = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    let maxValue = 2 ** 30;
    const QUARTER_OF_MAX_VALUE = maxValue / 4;
    let cnt = MAX_INLINE;
    continuously[cnt] = [maxValue, maxValue, maxValue];
    gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    if (MAX_INLINE > 1) {
        cnt -= 1;
        continuously[cnt] = [0, QUARTER_OF_MAX_VALUE, maxValue / 2];
        gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    }
    maxValue /= 8;
    for (; cnt > 0; --cnt, maxValue /= 8) {
        continuously[cnt] = [0, maxValue / 4, maxValue / 2];
        gap[cnt] = [0, maxValue / 8, maxValue / 4];
    }
    return { continuously, gap };
};

class GomokuSolution {
    MAX_DEPTH;
    context;
    state;
    stateCompressor;
    stateCache;
    scoreForPlayer;
    bestR;
    bestC;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE = 5, MAX_DEPTH = 3, scoreMap, NEXT_MOVER_FAC) {
        const context = new GomokuContext(MAX_ROW, MAX_COL, MAX_INLINE);
        const _scoreMap = scoreMap ?? createScoreMap(context.MAX_INLINE);
        this.MAX_DEPTH = Math.max(1, Math.round(MAX_DEPTH));
        this.context = context;
        this.state = new GomokuState(context, _scoreMap, NEXT_MOVER_FAC);
        this.stateCompressor = new GomokuStateCompressor(BigInt(context.TOTAL_POS));
        this.stateCache = new Map();
        this.scoreForPlayer = -1;
        this.bestR = -1;
        this.bestC = -1;
    }
    init(pieces) {
        this.state.init(pieces);
        this.stateCache.clear();
    }
    move(r, c, p) {
        this.state.forward(r, c, p);
    }
    minmaxMatch(currentPlayer) {
        if (this.state.isFinal())
            return { r: -1, c: -1 };
        this.stateCache.clear();
        this.scoreForPlayer = currentPlayer;
        this.bestR = this.bestC = -1;
        this.alphaBeta(currentPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, this.stateCompressor.INITIAL_STATE, 0);
        const { bestR: r, bestC: c } = this;
        return r < 0 || c < 0 ? this.state.randomMove() : { r, c };
    }
    alphaBeta(player, alpha, beta, cur, prevState, stateScore) {
        const { context, state, stateCompressor, stateCache, scoreForPlayer } = this;
        if (cur === this.MAX_DEPTH || state.isFinal())
            return stateScore;
        const candidates = state.expand(player, scoreForPlayer);
        if (candidates.length <= 0)
            return stateScore;
        if (player === scoreForPlayer) {
            candidates.sort((x, y) => y.score - x.score);
        }
        else {
            candidates.sort((x, y) => x.score - y.score);
        }
        if (cur === 0 && this.bestR === -1) {
            const { r, c } = candidates[0];
            this.bestR = r;
            this.bestC = c;
        }
        for (const { r, c, score } of candidates) {
            const id = context.idx(r, c);
            const nextState = stateCompressor.compress(cur, prevState, BigInt(id));
            let gamma = stateCache.get(nextState);
            if (gamma === undefined) {
                state.forward(r, c, player);
                gamma = this.alphaBeta(player ^ 1, alpha, beta, cur + 1, nextState, score);
                state.rollback(r, c);
                stateCache.set(nextState, gamma);
            }
            if (cur === 0) {
                if (alpha < gamma) {
                    this.bestR = r;
                    this.bestC = c;
                }
            }
            if (player === scoreForPlayer) {
                if (alpha < gamma)
                    alpha = gamma;
            }
            else {
                if (beta > gamma)
                    beta = gamma;
            }
            if (beta <= alpha)
                break;
        }
        return player === scoreForPlayer ? alpha : beta;
    }
}

export { GomokuContext, GomokuCountMap, GomokuDirectionType, GomokuSolution, GomokuState, GomokuStateCompressor, createScoreMap, gomokuDirectionTypes, gomokuDirections, leftHalfGomokuDirectionTypes, rightHalfGomokuDirectionTypes };
